@@ -1,6 +1,7 @@
 package com.fasterxml.jackson.core.io;
 
-import com.fasterxml.jackson.core.util.BufferRecyclers;
+import java.util.Arrays;
+
 import com.fasterxml.jackson.core.util.ByteArrayBuilder;
 import com.fasterxml.jackson.core.util.TextBuffer;
 
@@ -11,12 +12,15 @@ import com.fasterxml.jackson.core.util.TextBuffer;
  * Note that methods in here are somewhat optimized, but not ridiculously so.
  * Reason is that conversion method results are expected to be cached so that
  * these methods will not be hot spots during normal operation.
- *<p>
- * NOTE: starting with 2.9.3, access to most functionality should go through
- * {@link BufferRecyclers} and NOT directly through this class.
  */
 public final class JsonStringEncoder
 {
+    /*
+    /**********************************************************************
+    /* Constants
+    /**********************************************************************
+     */
+
     private final static char[] HC = CharTypes.copyHexChars();
 
     private final static byte[] HB = CharTypes.copyHexBytes();
@@ -26,74 +30,135 @@ public final class JsonStringEncoder
     private final static int SURR2_FIRST = 0xDC00;
     private final static int SURR2_LAST = 0xDFFF;
 
-//    private final static int INT_BACKSLASH = '\\';
-//    private final static int INT_U = 'u';
-//    private final static int INT_0 = '0';
+    private final static int INITIAL_CHAR_BUFFER_SIZE = 120;
+    private final static int INITIAL_BYTE_BUFFER_SIZE = 200;
 
-    /**
-     * Lazily constructed text buffer used to produce JSON encoded Strings
-     * as characters (without UTF-8 encoding)
-     */
-    protected TextBuffer _text;
-
-    /**
-     * Lazily-constructed builder used for UTF-8 encoding of text values
-     * (quoted and unquoted)
-     */
-    protected ByteArrayBuilder _bytes;
-    
-    /**
-     * Temporary buffer used for composing quote/escape sequences
-     */
-    protected final char[] _qbuf;
-    
     /*
-    /**********************************************************
+    /**********************************************************************
     /* Construction, instance access
-    /**********************************************************
+    /**********************************************************************
      */
+
+    // Since 2.10 we have stateless singleton and NO fancy ThreadLocal/SofRef caching!!!
+    private final static JsonStringEncoder instance = new JsonStringEncoder();
     
-    public JsonStringEncoder() {
-        _qbuf = new char[6];
-        _qbuf[0] = '\\';
-        _qbuf[2] = '0';
-        _qbuf[3] = '0';
-    }
-    
+    public JsonStringEncoder() { }
+
     /**
      * Factory method for getting an instance; this is either recycled per-thread instance,
      * or a newly constructed one.
      *
-     * @deprecated Since 2.9.2 use {@link BufferRecyclers#getJsonStringEncoder()} instead
+     * @return Static stateless encoder instance
      */
-    @Deprecated
     public static JsonStringEncoder getInstance() {
-        return BufferRecyclers.getJsonStringEncoder();
+        return instance;
     }
 
     /*
-    /**********************************************************
+    /**********************************************************************
     /* Public API
-    /**********************************************************
+    /**********************************************************************
      */
 
     /**
-     * Method that will quote text contents using JSON standard quoting,
-     * and return results as a character array
+     * Method that will escape text contents using JSON standard escaping,
+     * and return results as a character array.
+     *
+     * @param input Value String to process
+     *
+     * @return JSON-escaped String matching {@code input}
      */
     public char[] quoteAsString(String input)
     {
-        TextBuffer textBuffer = _text;
-        if (textBuffer == null) {
-            // no allocator; can add if we must, shouldn't need to
-            _text = textBuffer = new TextBuffer(null);
+        char[] outputBuffer = new char[INITIAL_CHAR_BUFFER_SIZE];
+        final int[] escCodes = CharTypes.get7BitOutputEscapes();
+        final int escCodeCount = escCodes.length;
+        int inPtr = 0;
+        final int inputLen = input.length();
+        TextBuffer textBuffer = null;
+        int outPtr = 0;
+        char[] qbuf = null;
+
+        outer:
+        while (inPtr < inputLen) {
+            tight_loop:
+            while (true) {
+                char c = input.charAt(inPtr);
+                if (c < escCodeCount && escCodes[c] != 0) {
+                    break tight_loop;
+                }
+                if (outPtr >= outputBuffer.length) {
+                    if (textBuffer == null) {
+                        textBuffer = TextBuffer.fromInitial(outputBuffer);
+                    }
+                    outputBuffer = textBuffer.finishCurrentSegment();
+                    outPtr = 0;
+                }
+                outputBuffer[outPtr++] = c;
+                if (++inPtr >= inputLen) {
+                    break outer;
+                }
+            }
+            // something to escape; 2 or 6-char variant? 
+            if (qbuf == null) {
+                qbuf = _qbuf();
+            }
+            char d = input.charAt(inPtr++);
+            int escCode = escCodes[d];
+            int length = (escCode < 0)
+                    ? _appendNumeric(d, qbuf)
+                    : _appendNamed(escCode, qbuf);
+                    ;
+            if ((outPtr + length) > outputBuffer.length) {
+                int first = outputBuffer.length - outPtr;
+                if (first > 0) {
+                    System.arraycopy(qbuf, 0, outputBuffer, outPtr, first);
+                }
+                if (textBuffer == null) {
+                    textBuffer = TextBuffer.fromInitial(outputBuffer);
+                }
+                outputBuffer = textBuffer.finishCurrentSegment();
+                int second = length - first;
+                System.arraycopy(qbuf, first, outputBuffer, 0, second);
+                outPtr = second;
+            } else {
+                System.arraycopy(qbuf, 0, outputBuffer, outPtr, length);
+                outPtr += length;
+            }
         }
-        char[] outputBuffer = textBuffer.emptyAndGetCurrentSegment();
+
+        if (textBuffer == null) {
+            return Arrays.copyOfRange(outputBuffer, 0, outPtr);
+        }
+        textBuffer.setCurrentLength(outPtr);
+        return textBuffer.contentsAsArray();
+    }
+
+    /**
+     * Overloaded variant of {@link #quoteAsString(String)}.
+     *
+     * @param input Value {@link CharSequence} to process
+     *
+     * @return JSON-escaped String matching {@code input}
+     *
+     * @since 2.10
+     */
+    public char[] quoteAsString(CharSequence input)
+    {
+        // 15-Aug-2019, tatu: Optimize common case as JIT can't get rid of overhead otherwise
+        if (input instanceof String) {
+            return quoteAsString((String) input);
+        }
+
+        TextBuffer textBuffer = null;
+
+        char[] outputBuffer = new char[INITIAL_CHAR_BUFFER_SIZE];
         final int[] escCodes = CharTypes.get7BitOutputEscapes();
         final int escCodeCount = escCodes.length;
         int inPtr = 0;
         final int inputLen = input.length();
         int outPtr = 0;
+        char[] qbuf = null;
  
         outer:
         while (inPtr < inputLen) {
@@ -104,6 +169,9 @@ public final class JsonStringEncoder
                     break tight_loop;
                 }
                 if (outPtr >= outputBuffer.length) {
+                    if (textBuffer == null) {
+                        textBuffer = TextBuffer.fromInitial(outputBuffer);
+                    }
                     outputBuffer = textBuffer.finishCurrentSegment();
                     outPtr = 0;
                 }
@@ -113,26 +181,36 @@ public final class JsonStringEncoder
                 }
             }
             // something to escape; 2 or 6-char variant? 
+            if (qbuf == null) {
+                qbuf = _qbuf();
+            }
             char d = input.charAt(inPtr++);
             int escCode = escCodes[d];
             int length = (escCode < 0)
-                    ? _appendNumeric(d, _qbuf)
-                    : _appendNamed(escCode, _qbuf);
+                    ? _appendNumeric(d, qbuf)
+                    : _appendNamed(escCode, qbuf);
                     ;
             if ((outPtr + length) > outputBuffer.length) {
                 int first = outputBuffer.length - outPtr;
                 if (first > 0) {
-                    System.arraycopy(_qbuf, 0, outputBuffer, outPtr, first);
+                    System.arraycopy(qbuf, 0, outputBuffer, outPtr, first);
+                }
+                if (textBuffer == null) {
+                    textBuffer = TextBuffer.fromInitial(outputBuffer);
                 }
                 outputBuffer = textBuffer.finishCurrentSegment();
                 int second = length - first;
-                System.arraycopy(_qbuf, first, outputBuffer, 0, second);
+                System.arraycopy(qbuf, first, outputBuffer, 0, second);
                 outPtr = second;
             } else {
-                System.arraycopy(_qbuf, 0, outputBuffer, outPtr, length);
+                System.arraycopy(qbuf, 0, outputBuffer, outPtr, length);
                 outPtr += length;
             }
         }
+
+        if (textBuffer == null) {
+            return Arrays.copyOfRange(outputBuffer, 0, outPtr);
+        }
         textBuffer.setCurrentLength(outPtr);
         return textBuffer.contentsAsArray();
     }
@@ -142,6 +220,9 @@ public final class JsonStringEncoder
      * and append results to a supplied {@link StringBuilder}.
      * Use this variant if you have e.g. a {@link StringBuilder} and want to avoid superfluous copying of it.
      *
+     * @param input Value {@link CharSequence} to process
+     * @param output {@link StringBuilder} to append escaped contents to
+     *
      * @since 2.8
      */
     public void quoteAsString(CharSequence input, StringBuilder output)
@@ -150,6 +231,7 @@ public final class JsonStringEncoder
         final int escCodeCount = escCodes.length;
         int inPtr = 0;
         final int inputLen = input.length();
+        char[] qbuf = null;
 
         outer:
         while (inPtr < inputLen) {
@@ -165,32 +247,35 @@ public final class JsonStringEncoder
                 }
             }
             // something to escape; 2 or 6-char variant?
+            if (qbuf == null) {
+                qbuf = _qbuf();
+            }
             char d = input.charAt(inPtr++);
             int escCode = escCodes[d];
             int length = (escCode < 0)
-                    ? _appendNumeric(d, _qbuf)
-                    : _appendNamed(escCode, _qbuf);
-                    ;
-            output.append(_qbuf, 0, length);
+                    ? _appendNumeric(d, qbuf)
+                    : _appendNamed(escCode, qbuf);
+            output.append(qbuf, 0, length);
         }
     }
 
     /**
-     * Will quote given JSON String value using standard quoting, encode
-     * results as UTF-8, and return result as a byte array.
+     * Method that will escape text contents using JSON standard escaping,
+     * encode resulting String as UTF-8 bytes
+     * and return results as a  byte array.
+     *
+     * @param text Value {@link String} to process
+     *
+     * @return UTF-8 encoded bytes of JSON-escaped {@code text}
      */
     @SuppressWarnings("resource")
     public byte[] quoteAsUTF8(String text)
     {
-        ByteArrayBuilder bb = _bytes;
-        if (bb == null) {
-            // no allocator; can add if we must, shouldn't need to
-            _bytes = bb = new ByteArrayBuilder(null);
-        }
         int inputPtr = 0;
         int inputEnd = text.length();
         int outputPtr = 0;
-        byte[] outputBuffer = bb.resetAndGetFirstSegment();
+        byte[] outputBuffer = new byte[INITIAL_BYTE_BUFFER_SIZE];
+        ByteArrayBuilder bb = null;
         
         main:
         while (inputPtr < inputEnd) {
@@ -203,6 +288,9 @@ public final class JsonStringEncoder
                     break inner_loop;
                 }
                 if (outputPtr >= outputBuffer.length) {
+                    if (bb == null) {
+                        bb = ByteArrayBuilder.fromInitial(outputBuffer, outputPtr);
+                    }
                     outputBuffer = bb.finishCurrentSegment();
                     outputPtr = 0;
                 }
@@ -210,7 +298,10 @@ public final class JsonStringEncoder
                 if (++inputPtr >= inputEnd) {
                     break main;
                 }
-            }                
+            }
+            if (bb == null) {
+                bb = ByteArrayBuilder.fromInitial(outputBuffer, outputPtr);
+            }
             if (outputPtr >= outputBuffer.length) {
                 outputBuffer = bb.finishCurrentSegment();
                 outputPtr = 0;
@@ -269,27 +360,30 @@ public final class JsonStringEncoder
             }
             outputBuffer[outputPtr++] = (byte) ch;
         }
-        return _bytes.completeAndCoalesce(outputPtr);
+        if (bb == null) {
+            return Arrays.copyOfRange(outputBuffer, 0, outputPtr);
+        }
+        return bb.completeAndCoalesce(outputPtr);
     }
-    
+
     /**
-     * Will encode given String as UTF-8 (without any quoting), return
-     * resulting byte array.
+     * Will encode given String as UTF-8 (without any escaping) and return
+     * the resulting byte array.
+     *
+     * @param text Value {@link String} to process
+     *
+     * @return UTF-8 encoded bytes of {@code text} (without any escaping)
      */
     @SuppressWarnings("resource")
     public byte[] encodeAsUTF8(String text)
     {
-        ByteArrayBuilder byteBuilder = _bytes;
-        if (byteBuilder == null) {
-            // no allocator; can add if we must, shouldn't need to
-            _bytes = byteBuilder = new ByteArrayBuilder(null);
-        }
         int inputPtr = 0;
         int inputEnd = text.length();
         int outputPtr = 0;
-        byte[] outputBuffer = byteBuilder.resetAndGetFirstSegment();
+        byte[] outputBuffer = new byte[INITIAL_BYTE_BUFFER_SIZE];
         int outputEnd = outputBuffer.length;
-        
+        ByteArrayBuilder bb = null;
+
         main_loop:
         while (inputPtr < inputEnd) {
             int c = text.charAt(inputPtr++);
@@ -297,7 +391,10 @@ public final class JsonStringEncoder
             // first tight loop for ascii
             while (c <= 0x7F) {
                 if (outputPtr >= outputEnd) {
-                    outputBuffer = byteBuilder.finishCurrentSegment();
+                    if (bb == null) {
+                        bb = ByteArrayBuilder.fromInitial(outputBuffer, outputPtr);
+                    }
+                    outputBuffer = bb.finishCurrentSegment();
                     outputEnd = outputBuffer.length;
                     outputPtr = 0;
                 }
@@ -309,8 +406,11 @@ public final class JsonStringEncoder
             }
 
             // then multi-byte...
+            if (bb == null) {
+                bb = ByteArrayBuilder.fromInitial(outputBuffer, outputPtr);
+            }
             if (outputPtr >= outputEnd) {
-                outputBuffer = byteBuilder.finishCurrentSegment();
+                outputBuffer = bb.finishCurrentSegment();
                 outputEnd = outputBuffer.length;
                 outputPtr = 0;
             }
@@ -321,7 +421,7 @@ public final class JsonStringEncoder
                 if (c < SURR1_FIRST || c > SURR2_LAST) { // nope
                     outputBuffer[outputPtr++] = (byte) (0xe0 | (c >> 12));
                     if (outputPtr >= outputEnd) {
-                        outputBuffer = byteBuilder.finishCurrentSegment();
+                        outputBuffer = bb.finishCurrentSegment();
                         outputEnd = outputBuffer.length;
                         outputPtr = 0;
                     }
@@ -340,13 +440,13 @@ public final class JsonStringEncoder
                     }
                     outputBuffer[outputPtr++] = (byte) (0xf0 | (c >> 18));
                     if (outputPtr >= outputEnd) {
-                        outputBuffer = byteBuilder.finishCurrentSegment();
+                        outputBuffer = bb.finishCurrentSegment();
                         outputEnd = outputBuffer.length;
                         outputPtr = 0;
                     }
                     outputBuffer[outputPtr++] = (byte) (0x80 | ((c >> 12) & 0x3f));
                     if (outputPtr >= outputEnd) {
-                        outputBuffer = byteBuilder.finishCurrentSegment();
+                        outputBuffer = bb.finishCurrentSegment();
                         outputEnd = outputBuffer.length;
                         outputPtr = 0;
                     }
@@ -354,20 +454,132 @@ public final class JsonStringEncoder
                 }
             }
             if (outputPtr >= outputEnd) {
-                outputBuffer = byteBuilder.finishCurrentSegment();
+                outputBuffer = bb.finishCurrentSegment();
                 outputEnd = outputBuffer.length;
                 outputPtr = 0;
             }
             outputBuffer[outputPtr++] = (byte) (0x80 | (c & 0x3f));
         }
-        return _bytes.completeAndCoalesce(outputPtr);
+        if (bb == null) {
+            return Arrays.copyOfRange(outputBuffer, 0, outputPtr);
+        }
+        return bb.completeAndCoalesce(outputPtr);
     }
-    
-    /*
-    /**********************************************************
-    /* Internal methods
-    /**********************************************************
+
+    /**
+     * Overloaded variant of {@link #encodeAsUTF8(String)}.
+     *
+     * @param text Value {@link CharSequence} to process
+     *
+     * @return UTF-8 encoded bytes of {@code text} (without any escaping)
+     *
+     * @since 2.11
      */
+    @SuppressWarnings("resource")
+    public byte[] encodeAsUTF8(CharSequence text)
+    {
+        int inputPtr = 0;
+        int inputEnd = text.length();
+        int outputPtr = 0;
+        byte[] outputBuffer = new byte[INITIAL_BYTE_BUFFER_SIZE];
+        int outputEnd = outputBuffer.length;
+        ByteArrayBuilder bb = null;
+
+        main_loop:
+        while (inputPtr < inputEnd) {
+            int c = text.charAt(inputPtr++);
+
+            // first tight loop for ascii
+            while (c <= 0x7F) {
+                if (outputPtr >= outputEnd) {
+                    if (bb == null) {
+                        bb = ByteArrayBuilder.fromInitial(outputBuffer, outputPtr);
+                    }
+                    outputBuffer = bb.finishCurrentSegment();
+                    outputEnd = outputBuffer.length;
+                    outputPtr = 0;
+                }
+                outputBuffer[outputPtr++] = (byte) c;
+                if (inputPtr >= inputEnd) {
+                    break main_loop;
+                }
+                c = text.charAt(inputPtr++);
+            }
+
+            // then multi-byte...
+            if (bb == null) {
+                bb = ByteArrayBuilder.fromInitial(outputBuffer, outputPtr);
+            }
+            if (outputPtr >= outputEnd) {
+                outputBuffer = bb.finishCurrentSegment();
+                outputEnd = outputBuffer.length;
+                outputPtr = 0;
+            }
+            if (c < 0x800) { // 2-byte
+                outputBuffer[outputPtr++] = (byte) (0xc0 | (c >> 6));
+            } else { // 3 or 4 bytes
+                // Surrogates?
+                if (c < SURR1_FIRST || c > SURR2_LAST) { // nope
+                    outputBuffer[outputPtr++] = (byte) (0xe0 | (c >> 12));
+                    if (outputPtr >= outputEnd) {
+                        outputBuffer = bb.finishCurrentSegment();
+                        outputEnd = outputBuffer.length;
+                        outputPtr = 0;
+                    }
+                    outputBuffer[outputPtr++] = (byte) (0x80 | ((c >> 6) & 0x3f));
+                } else { // yes, surrogate pair
+                    if (c > SURR1_LAST) { // must be from first range
+                        _illegal(c);
+                    }
+                    // and if so, followed by another from next range
+                    if (inputPtr >= inputEnd) {
+                        _illegal(c);
+                    }
+                    c = _convert(c, text.charAt(inputPtr++));
+                    if (c > 0x10FFFF) { // illegal, as per RFC 4627
+                        _illegal(c);
+                    }
+                    outputBuffer[outputPtr++] = (byte) (0xf0 | (c >> 18));
+                    if (outputPtr >= outputEnd) {
+                        outputBuffer = bb.finishCurrentSegment();
+                        outputEnd = outputBuffer.length;
+                        outputPtr = 0;
+                    }
+                    outputBuffer[outputPtr++] = (byte) (0x80 | ((c >> 12) & 0x3f));
+                    if (outputPtr >= outputEnd) {
+                        outputBuffer = bb.finishCurrentSegment();
+                        outputEnd = outputBuffer.length;
+                        outputPtr = 0;
+                    }
+                    outputBuffer[outputPtr++] = (byte) (0x80 | ((c >> 6) & 0x3f));
+                }
+            }
+            if (outputPtr >= outputEnd) {
+                outputBuffer = bb.finishCurrentSegment();
+                outputEnd = outputBuffer.length;
+                outputPtr = 0;
+            }
+            outputBuffer[outputPtr++] = (byte) (0x80 | (c & 0x3f));
+        }
+        if (bb == null) {
+            return Arrays.copyOfRange(outputBuffer, 0, outputPtr);
+        }
+        return bb.completeAndCoalesce(outputPtr);
+    }
+
+    /*
+    /**********************************************************************
+    /* Internal methods
+    /**********************************************************************
+     */
+
+    private char[] _qbuf() {
+        char[] qbuf = new char[6];
+        qbuf[0] = '\\';
+        qbuf[2] = '0';
+        qbuf[3] = '0';
+        return qbuf;
+    }
 
     private int _appendNumeric(int value, char[] qbuf) {
         qbuf[1] = 'u';
