@@ -5,13 +5,16 @@ import java.math.BigDecimal;
 import java.math.BigInteger;
 
 import com.fasterxml.jackson.core.*;
-import com.fasterxml.jackson.core.io.*;
+import com.fasterxml.jackson.core.io.CharTypes;
+import com.fasterxml.jackson.core.io.CharacterEscapes;
+import com.fasterxml.jackson.core.io.IOContext;
+import com.fasterxml.jackson.core.io.NumberOutput;
 
 /**
  * {@link JsonGenerator} that outputs JSON content using a {@link java.io.Writer}
  * which handles character encoding.
  */
-public final class WriterBasedJsonGenerator
+public class WriterBasedJsonGenerator
     extends JsonGeneratorImpl
 {
     final protected static int SHORT_WRITE = 32;
@@ -29,10 +32,8 @@ public final class WriterBasedJsonGenerator
     /**
      * Character used for quoting JSON Object property names
      * and String values.
-     *
-     * @since 2.8
      */
-    protected char _quoteChar = '"'; // TODO: make configurable
+    protected char _quoteChar;
 
     /*
     /**********************************************************
@@ -75,27 +76,50 @@ public final class WriterBasedJsonGenerator
      */
     protected SerializableString _currentEscape;
 
+    /**
+     * Intermediate buffer in which characters of a String are copied
+     * before being encoded.
+     *
+     * @since 2.10
+     */
+    protected char[] _copyBuffer;
+
     /*
     /**********************************************************
     /* Life-cycle
     /**********************************************************
      */
 
+    @Deprecated // since 2.10
     public WriterBasedJsonGenerator(IOContext ctxt, int features,
-            ObjectCodec codec, Writer w)
+            ObjectCodec codec, Writer w) {
+        this(ctxt, features, codec, w, JsonFactory.DEFAULT_QUOTE_CHAR);
+    }
+
+    /**
+     * @since 2.10
+     */
+    public WriterBasedJsonGenerator(IOContext ctxt, int features,
+            ObjectCodec codec, Writer w,
+            char quoteChar)
+    
     {
         super(ctxt, features, codec);
         _writer = w;
         _outputBuffer = ctxt.allocConcatBuffer();
         _outputEnd = _outputBuffer.length;
+        _quoteChar = quoteChar;
+        if (quoteChar != '"') { // since 2.10
+            _outputEscapes = CharTypes.get7BitOutputEscapes(quoteChar);
+        }
     }
-    
+
     /*
     /**********************************************************
     /* Overridden configuration, introspection methods
     /**********************************************************
      */
-    
+
     @Override
     public Object getOutputTarget() {
         return _writer;
@@ -139,7 +163,7 @@ public final class WriterBasedJsonGenerator
         _writeFieldName(name, (status == JsonWriteContext.STATUS_OK_AFTER_COMMA));
     }
 
-    protected void _writeFieldName(String name, boolean commaBefore) throws IOException
+    protected final void _writeFieldName(String name, boolean commaBefore) throws IOException
     {
         if (_cfgPrettyPrinter != null) {
             _writePPFieldName(name, commaBefore);
@@ -168,7 +192,7 @@ public final class WriterBasedJsonGenerator
         _outputBuffer[_outputTail++] = _quoteChar;
     }
 
-    protected void _writeFieldName(SerializableString name, boolean commaBefore) throws IOException
+    protected final void _writeFieldName(SerializableString name, boolean commaBefore) throws IOException
     {
         if (_cfgPrettyPrinter != null) {
             _writePPFieldName(name, commaBefore);
@@ -182,29 +206,37 @@ public final class WriterBasedJsonGenerator
             _outputBuffer[_outputTail++] = ',';
         }
         // Alternate mode, in which quoting of field names disabled?
-        final char[] quoted = name.asQuotedChars();
         if (_cfgUnqNames) {
-            writeRaw(quoted, 0, quoted.length);
+            final char[] ch = name.asQuotedChars();
+            writeRaw(ch, 0, ch.length);
             return;
         }
         // we know there's room for at least one more char
         _outputBuffer[_outputTail++] = _quoteChar;
         // The beef:
-        final int qlen = quoted.length;
-        if ((_outputTail + qlen + 1) >= _outputEnd) {
-            writeRaw(quoted, 0, qlen);
-            // and closing quotes; need room for one more char:
-            if (_outputTail >= _outputEnd) {
-                _flushBuffer();
-            }
-            _outputBuffer[_outputTail++] = _quoteChar;
-        } else {
-            System.arraycopy(quoted, 0, _outputBuffer, _outputTail, qlen);
-            _outputTail += qlen;
-            _outputBuffer[_outputTail++] = _quoteChar;
+        
+        int len = name.appendQuoted(_outputBuffer, _outputTail);
+        if (len < 0) {
+            _writeFieldNameTail(name);
+            return;
         }
+        _outputTail += len;
+        if (_outputTail >= _outputEnd) {
+            _flushBuffer();
+        }
+        _outputBuffer[_outputTail++] = _quoteChar;
     }
-    
+
+    private final void _writeFieldNameTail(SerializableString name) throws IOException
+    {
+        final char[] quoted = name.asQuotedChars();
+        writeRaw(quoted, 0, quoted.length);
+        if (_outputTail >= _outputEnd) {
+            _flushBuffer();
+        }
+        _outputBuffer[_outputTail++] = _quoteChar;
+    }
+
     /*
     /**********************************************************
     /* Output method implementations, structural
@@ -226,6 +258,36 @@ public final class WriterBasedJsonGenerator
         }
     }
 
+    @Override // since 2.12
+    public void writeStartArray(Object currentValue) throws IOException
+    {
+        _verifyValueWrite("start an array");
+        _writeContext = _writeContext.createChildArrayContext(currentValue);
+        if (_cfgPrettyPrinter != null) {
+            _cfgPrettyPrinter.writeStartArray(this);
+        } else {
+            if (_outputTail >= _outputEnd) {
+                _flushBuffer();
+            }
+            _outputBuffer[_outputTail++] = '[';
+        }
+    }
+
+    @Override // since 2.12
+    public void writeStartArray(Object currentValue, int size) throws IOException
+    {
+        _verifyValueWrite("start an array");
+        _writeContext = _writeContext.createChildArrayContext(currentValue);
+        if (_cfgPrettyPrinter != null) {
+            _cfgPrettyPrinter.writeStartArray(this);
+        } else {
+            if (_outputTail >= _outputEnd) {
+                _flushBuffer();
+            }
+            _outputBuffer[_outputTail++] = '[';
+        }
+    }
+
     @Override
     public void writeEndArray() throws IOException
     {
@@ -243,15 +305,11 @@ public final class WriterBasedJsonGenerator
         _writeContext = _writeContext.clearAndGetParent();
     }
 
-    @Override // since 2.8
-    public void writeStartObject(Object forValue) throws IOException
+    @Override
+    public void writeStartObject() throws IOException
     {
         _verifyValueWrite("start an object");
-        JsonWriteContext ctxt = _writeContext.createChildObjectContext();
-        _writeContext = ctxt;
-        if (forValue != null) {
-            ctxt.setCurrentValue(forValue);
-        }
+        _writeContext = _writeContext.createChildObjectContext();
         if (_cfgPrettyPrinter != null) {
             _cfgPrettyPrinter.writeStartObject(this);
         } else {
@@ -261,12 +319,13 @@ public final class WriterBasedJsonGenerator
             _outputBuffer[_outputTail++] = '{';
         }
     }
-    
-    @Override
-    public void writeStartObject() throws IOException
+
+    @Override // since 2.8
+    public void writeStartObject(Object forValue) throws IOException
     {
         _verifyValueWrite("start an object");
-        _writeContext = _writeContext.createChildObjectContext();
+        JsonWriteContext ctxt = _writeContext.createChildObjectContext(forValue);
+        _writeContext = ctxt;
         if (_cfgPrettyPrinter != null) {
             _cfgPrettyPrinter.writeStartObject(this);
         } else {
@@ -298,7 +357,7 @@ public final class WriterBasedJsonGenerator
      * Specialized version of <code>_writeFieldName</code>, off-lined
      * to keep the "fast path" as simple (and hopefully fast) as possible.
      */
-    protected void _writePPFieldName(String name, boolean commaBefore) throws IOException
+    protected final void _writePPFieldName(String name, boolean commaBefore) throws IOException
     {
         if (commaBefore) {
             _cfgPrettyPrinter.writeObjectEntrySeparator(this);
@@ -321,14 +380,13 @@ public final class WriterBasedJsonGenerator
         }
     }
 
-    protected void _writePPFieldName(SerializableString name, boolean commaBefore) throws IOException
+    protected final void _writePPFieldName(SerializableString name, boolean commaBefore) throws IOException
     {
         if (commaBefore) {
             _cfgPrettyPrinter.writeObjectEntrySeparator(this);
         } else {
             _cfgPrettyPrinter.beforeObjectEntries(this);
         }
-    
         final char[] quoted = name.asQuotedChars();
         if (_cfgUnqNames) {// non-standard, omit quotes
             writeRaw(quoted, 0, quoted.length);
@@ -371,6 +429,47 @@ public final class WriterBasedJsonGenerator
         _outputBuffer[_outputTail++] = _quoteChar;
     }
 
+    @Override
+    public void writeString(Reader reader, int len) throws IOException {
+        _verifyValueWrite(WRITE_STRING);
+        if (reader == null) {
+            _reportError("null reader");
+            return; // just to block warnings by lgtm.com
+        }
+        int toRead = (len >= 0) ? len : Integer.MAX_VALUE;
+        //Add leading quote
+        if ((_outputTail + len) >= _outputEnd) {
+            _flushBuffer();
+        }
+        _outputBuffer[_outputTail++] = _quoteChar;
+
+        final char[] buf = _allocateCopyBuffer();
+        //read
+        while (toRead > 0) {
+            int toReadNow = Math.min(toRead, buf.length);
+            int numRead = reader.read(buf, 0, toReadNow);
+            if (numRead <= 0) {
+                break;
+            }
+            if ((_outputTail + len) >= _outputEnd) {
+                _flushBuffer();
+            }
+            _writeString(buf, 0, numRead);
+
+            //decrease tracker
+            toRead -= numRead;
+        }
+        //Add trailing quote
+        if ((_outputTail + len) >= _outputEnd) {
+            _flushBuffer();
+        }
+        _outputBuffer[_outputTail++] = _quoteChar;
+
+        if (toRead > 0 && len >= 0) {
+            _reportError("Didn't read enough from reader");
+        }
+    }
+
     @Override
     public void writeString(char[] text, int offset, int len) throws IOException
     {
@@ -395,10 +494,23 @@ public final class WriterBasedJsonGenerator
             _flushBuffer();
         }
         _outputBuffer[_outputTail++] = _quoteChar;
+        int len = sstr.appendQuoted(_outputBuffer, _outputTail);
+        if (len < 0) {
+            _writeString2(sstr);
+            return;
+        }
+        _outputTail += len;
+        if (_outputTail >= _outputEnd) {
+            _flushBuffer();
+        }
+        _outputBuffer[_outputTail++] = _quoteChar;
+    }
+
+    private void _writeString2(SerializableString sstr) throws IOException
+    {
         // Note: copied from writeRaw:
         char[] text = sstr.asQuotedChars();
         final int len = text.length;
-        // Only worth buffering if it's a short write?
         if (len < SHORT_WRITE) {
             int room = _outputEnd - _outputTail;
             if (len > room) {
@@ -407,7 +519,6 @@ public final class WriterBasedJsonGenerator
             System.arraycopy(text, 0, _outputBuffer, _outputTail, len);
             _outputTail += len;
         } else {
-            // Otherwise, better just pass through:
             _flushBuffer();
             _writer.write(text, 0, len);
         }
@@ -416,7 +527,7 @@ public final class WriterBasedJsonGenerator
         }
         _outputBuffer[_outputTail++] = _quoteChar;
     }
-
+    
     @Override
     public void writeRawUTF8String(byte[] text, int offset, int length) throws IOException {
         // could add support for buffering if we really want it...
@@ -477,7 +588,12 @@ public final class WriterBasedJsonGenerator
     // @since 2.1
     @Override
     public void writeRaw(SerializableString text) throws IOException {
-        writeRaw(text.getValue());
+        int len = text.appendUnquoted(_outputBuffer, _outputTail);
+        if (len < 0) {
+            writeRaw(text.getValue());
+            return;
+        }
+        _outputTail += len;
     }
 
     @Override
@@ -683,13 +799,12 @@ public final class WriterBasedJsonGenerator
         }
     }
 
-    
+    @SuppressWarnings("deprecation")
     @Override
     public void writeNumber(double d) throws IOException
     {
         if (_cfgNumbersAsStrings ||
-            // [JACKSON-139]
-                (isEnabled(Feature.QUOTE_NON_NUMERIC_NUMBERS) && ((Double.isNaN(d) || Double.isInfinite(d))))) {
+                (NumberOutput.notFinite(d) && isEnabled(Feature.QUOTE_NON_NUMERIC_NUMBERS))) {
             writeString(String.valueOf(d));
             return;
         }
@@ -698,12 +813,12 @@ public final class WriterBasedJsonGenerator
         writeRaw(String.valueOf(d));
     }
 
+    @SuppressWarnings("deprecation")
     @Override
     public void writeNumber(float f) throws IOException
     {
         if (_cfgNumbersAsStrings ||
-            // [JACKSON-139]
-                (isEnabled(Feature.QUOTE_NON_NUMERIC_NUMBERS) && ((Float.isNaN(f) || Float.isInfinite(f))))) {
+                (NumberOutput.notFinite(f) && isEnabled(Feature.QUOTE_NON_NUMERIC_NUMBERS))) {
             writeString(String.valueOf(f));
             return;
         }
@@ -730,13 +845,25 @@ public final class WriterBasedJsonGenerator
     public void writeNumber(String encodedValue) throws IOException
     {
         _verifyValueWrite(WRITE_NUMBER);
-        if (_cfgNumbersAsStrings) {
+        if (encodedValue == null) {
+            _writeNull();
+        } else if (_cfgNumbersAsStrings) {
             _writeQuotedRaw(encodedValue);
         } else {
             writeRaw(encodedValue);
         }
     }
 
+    @Override
+    public void writeNumber(char[] encodedValueBuffer, int offset, int length) throws IOException {
+        _verifyValueWrite(WRITE_NUMBER);
+        if (_cfgNumbersAsStrings) {
+            _writeQuotedRaw(encodedValueBuffer, offset, length);
+        } else {
+            writeRaw(encodedValueBuffer, offset, length);
+        }
+    }
+
     private void _writeQuotedRaw(String value) throws IOException
     {
         if (_outputTail >= _outputEnd) {
@@ -750,6 +877,19 @@ public final class WriterBasedJsonGenerator
         _outputBuffer[_outputTail++] = _quoteChar;
     }
 
+    private void _writeQuotedRaw(char[] text, int offset, int length) throws IOException
+    {
+        if (_outputTail >= _outputEnd) {
+            _flushBuffer();
+        }
+        _outputBuffer[_outputTail++] = _quoteChar;
+        writeRaw(text, offset, length);
+        if (_outputTail >= _outputEnd) {
+            _flushBuffer();
+        }
+        _outputBuffer[_outputTail++] = _quoteChar;
+    }
+
     @Override
     public void writeBoolean(boolean state) throws IOException
     {
@@ -787,7 +927,7 @@ public final class WriterBasedJsonGenerator
      */
 
     @Override
-    protected void _verifyValueWrite(String typeMsg) throws IOException
+    protected final void _verifyValueWrite(String typeMsg) throws IOException
     {
         final int status = _writeContext.writeValue();
         if (_cfgPrettyPrinter != null) {
@@ -890,6 +1030,11 @@ public final class WriterBasedJsonGenerator
             _outputBuffer = null;
             _ioContext.releaseConcatBuffer(buf);
         }
+        buf = _copyBuffer;
+        if (buf != null) {
+            _copyBuffer = null;
+            _ioContext.releaseNameCopyBuffer(buf);
+        }
     }
 
     /*
@@ -1054,11 +1199,10 @@ public final class WriterBasedJsonGenerator
             _writeStringASCII(text, offset, len, _maximumNonEscapedChar);
             return;
         }
-        
-        /* Let's just find longest spans of non-escapable
-         * content, and for each see if it makes sense
-         * to copy them, or write through
-         */
+
+        // Let's just find longest spans of non-escapable content, and for
+        // each see if it makes sense to copy them, or write through
+
         len += offset; // -> len marks the end from now on
         final int[] escCodes = _outputEscapes;
         final int escLen = escCodes.length;
@@ -1414,7 +1558,7 @@ public final class WriterBasedJsonGenerator
     /**********************************************************
      */
     
-    protected void _writeBinary(Base64Variant b64variant, byte[] input, int inputPtr, final int inputEnd)
+    protected final void _writeBinary(Base64Variant b64variant, byte[] input, int inputPtr, final int inputEnd)
         throws IOException, JsonGenerationException
     {
         // Encoding is by chunks of 3 input, 4 output chars, so:
@@ -1456,7 +1600,7 @@ public final class WriterBasedJsonGenerator
     }
 
     // write-method called when length is definitely known
-    protected int _writeBinary(Base64Variant b64variant,
+    protected final int _writeBinary(Base64Variant b64variant,
             InputStream data, byte[] readBuffer, int bytesLeft)
         throws IOException, JsonGenerationException
     {
@@ -1516,7 +1660,7 @@ public final class WriterBasedJsonGenerator
     }
     
     // write method when length is unknown
-    protected int _writeBinary(Base64Variant b64variant,
+    protected final int _writeBinary(Base64Variant b64variant,
             InputStream data, byte[] readBuffer)
         throws IOException, JsonGenerationException
     {
@@ -1870,7 +2014,17 @@ public final class WriterBasedJsonGenerator
         _entityBuffer = buf;
         return buf;
     }
-    
+
+    /**
+     * @since 2.9
+     */
+    private char[] _allocateCopyBuffer() {
+        if (_copyBuffer == null) {
+            _copyBuffer = _ioContext.allocNameCopyBuffer(2000);
+        }
+        return _copyBuffer;
+    }
+
     protected void _flushBuffer() throws IOException
     {
         int len = _outputTail - _outputHead;
