@@ -11,6 +11,18 @@ import com.fasterxml.jackson.core.util.InternCache;
  * memory access due to flattening of name quad data.
  * Performance improvement modest for simple JSON document data binding (maybe 3%),
  * but should help more for larger symbol tables, or for binary formats like Smile.
+ *<p>
+ * Hash area is divided into 4 sections:
+ *<ol>
+ * <li>Primary area (1/2 of total size), direct match from hash (LSB)</li>
+ * <li>Secondary area (1/4 of total size), match from {@code hash (LSB) >> 1}</li>
+ * <li>Tertiary area (1/8 of total size), match from {@code hash (LSB) >> 2}</li>
+ * <li>Spill-over area (remaining 1/8) with linear scan, insertion order</li>
+ * <li></li>
+ * </ol>
+ * and within every area, entries are 4 {@code int}s, where 1 - 3 ints contain 1 - 12
+ * UTF-8 encoded bytes of name (null-padded), and last int is offset in
+ * {@code _names} that contains actual name Strings.
  *
  * @since 2.6
  */
@@ -18,7 +30,7 @@ public final class ByteQuadsCanonicalizer
 {
     /**
      * Initial size of the primary hash area. Each entry consumes 4 ints (16 bytes),
-     * and secondary area is same as primary; so default size will use 2kB of memory_tertiaryStart
+     * and secondary area is same as primary; so default size will use 2kB of memory
      * (plus 64x4 or 64x8 (256/512 bytes) for references to Strings, and Strings
      * themselves).
      */
@@ -27,7 +39,6 @@ public final class ByteQuadsCanonicalizer
 
     /**
      * Let's not expand symbol tables past some maximum size;
-     * this should protected against OOMEs caused by large documents
      * with unique (~= random) names.
      * Size is in 
      */
@@ -56,15 +67,15 @@ public final class ByteQuadsCanonicalizer
      * Reference to the root symbol table, for child tables, so
      * that they can merge table information back as necessary.
      */
-    final private ByteQuadsCanonicalizer _parent;
+    final protected ByteQuadsCanonicalizer _parent;
 
     /**
      * Member that is only used by the root table instance: root
-     * passes immutable state into child instances, and children
+     * passes immutable state info child instances, and children
      * may return new state if they add entries to the table.
      * Child tables do NOT use the reference.
      */
-    final private AtomicReference<TableInfo> _tableInfo;
+    final protected AtomicReference<TableInfo> _tableInfo;
     
     /**
      * Seed value we use as the base to make hash codes non-static between
@@ -73,7 +84,7 @@ public final class ByteQuadsCanonicalizer
      * This is done for security reasons, to avoid potential DoS attack via
      * hash collisions.
      */
-    final private int _seed;
+    final protected int _seed;
     
     /*
     /**********************************************************
@@ -88,7 +99,7 @@ public final class ByteQuadsCanonicalizer
      * NOTE: non-final to allow disabling intern()ing in case of excessive
      * collisions.
      */
-    private boolean _intern;
+    protected boolean _intern;
 
     /**
      * Flag that indicates whether we should throw an exception if enough 
@@ -96,7 +107,7 @@ public final class ByteQuadsCanonicalizer
      * 
      * @since 2.4
      */
-    private final boolean _failOnDoS;
+    protected final boolean _failOnDoS;
     
     /*
     /**********************************************************
@@ -110,7 +121,7 @@ public final class ByteQuadsCanonicalizer
      * structure (details of which may be tweaked depending on expected rates
      * of collisions).
      */
-    private int[] _hashArea;
+    protected int[] _hashArea;
 
     /**
      * Number of slots for primary entries within {@link #_hashArea}; which is
@@ -118,17 +129,17 @@ public final class ByteQuadsCanonicalizer
      * primary covers only half of the area; plus, additional area for longer
      * symbols after hash area).
      */
-    private int _hashSize;
+    protected int _hashSize;
 
     /**
      * Offset within {@link #_hashArea} where secondary entries start
      */
-    private int _secondaryStart;
+    protected int _secondaryStart;
 
     /**
      * Offset within {@link #_hashArea} where tertiary entries start
      */
-    private int _tertiaryStart;
+    protected int _tertiaryStart;
     
     /**
      * Constant that determines size of buckets for tertiary entries:
@@ -139,12 +150,12 @@ public final class ByteQuadsCanonicalizer
      * Default value is 2, for buckets of 4 slots; grows bigger with
      * bigger table sizes.
      */
-    private int _tertiaryShift;
+    protected int _tertiaryShift;
 
     /**
      * Total number of Strings in the symbol table; only used for child tables.
      */
-    private int _count;
+    protected int _count;
 
     /**
      * Array that contains <code>String</code> instances matching
@@ -152,7 +163,7 @@ public final class ByteQuadsCanonicalizer
      * Contains nulls for unused entries. Note that this size is twice
      * that of {@link #_hashArea}
      */
-    private String[] _names;
+    protected String[] _names;
 
     /*
     /**********************************************************
@@ -165,23 +176,17 @@ public final class ByteQuadsCanonicalizer
      * for more spilled over entries (if any).
      * Spill over area is within fixed-size portion of {@link #_hashArea}.
      */
-    private int _spilloverEnd;
+    protected int _spilloverEnd;
 
     /**
      * Offset within {@link #_hashArea} that follows main slots and contains
-     * quads for longer names (13 bytes or longers), and points to the
+     * quads for longer names (13 bytes or longer), and points to the
      * first available int that may be used for appending quads of the next
      * long name.
      * Note that long name area follows immediately after the fixed-size
      * main hash area ({@link #_hashArea}).
      */
-    private int _longNameOffset;
-
-    /**
-     * This flag is set if, after adding a new entry, it is deemed
-     * that a rehash is warranted if any more entries are to be added.
-     */
-    private transient boolean _needRehash;
+    protected int _longNameOffset;
 
     /*
     /**********************************************************
@@ -202,7 +207,7 @@ public final class ByteQuadsCanonicalizer
      * and when adding new collision list queues (i.e. creating a new
      * collision list head entry)
      */
-    private boolean _hashShared;
+    protected boolean _hashShared;
 
     /*
     /**********************************************************
@@ -267,7 +272,6 @@ public final class ByteQuadsCanonicalizer
         _longNameOffset = state.longNameOffset;
 
         // and then set other state to reflect sharing status
-        _needRehash = false;
         _hashShared = true;
     }
 
@@ -280,28 +284,31 @@ public final class ByteQuadsCanonicalizer
     /**
      * Factory method to call to create a symbol table instance with a
      * randomized seed value.
+     *
+     * @return Root instance to use for constructing new child instances 
      */
     public static ByteQuadsCanonicalizer createRoot() {
-        /* [Issue-21]: Need to use a variable seed, to thwart hash-collision
-         * based attacks.
-         */
+        // Need to use a variable seed, to thwart hash-collision based attacks.
+        // 14-Feb-2017, tatu: Does this actually help?
         long now = System.currentTimeMillis();
         // ensure it's not 0; and might as well require to be odd so:
         int seed = (((int) now) + ((int) (now >>> 32))) | 1;
         return createRoot(seed);
     }
 
-    /**
-     * Factory method that should only be called from unit tests, where seed
-     * value should remain the same.
-     */
+    // Factory method that should only be called from unit tests, where seed
+    // value should remain the same.
     protected static ByteQuadsCanonicalizer createRoot(int seed) {
         return new ByteQuadsCanonicalizer(DEFAULT_T_SIZE, true, seed, true);
     }
-    
+
     /**
      * Factory method used to create actual symbol table instance to
      * use for parsing.
+     *
+     * @param flags Bit flags of active {@link com.fasterxml.jackson.core.JsonFactory.Feature}s enabled.
+     *
+     * @return Actual canonicalizer instance that can be used by a parser
      */
     public ByteQuadsCanonicalizer makeChild(int flags) {
         return new ByteQuadsCanonicalizer(this,
@@ -312,20 +319,19 @@ public final class ByteQuadsCanonicalizer
     }
 
     /**
-     * Method called by the using code to indicate it is done
-     * with this instance. This lets instance merge accumulated
-     * changes into parent (if need be), safely and efficiently,
-     * and without calling code having to know about parent
-     * information
+     * Method called by the using code to indicate it is done with this instance.
+     * This lets instance merge accumulated changes into parent (if need be),
+     * safely and efficiently, and without calling code having to know about parent
+     * information.
      */
     public void release()
     {
         // we will try to merge if child table has new entries
-        if (_parent != null && maybeDirty()) {
+        // 28-Jul-2019, tatu: From [core#548]: do not share if immediate rehash needed
+        if ((_parent != null) && maybeDirty()) {
             _parent.mergeChild(new TableInfo(this));
-            /* Let's also mark this instance as dirty, so that just in
-             * case release was too early, there's no corruption of possibly shared data.
-             */
+            // Let's also mark this instance as dirty, so that just in
+            // case release was too early, there's no corruption of possibly shared data.
             _hashShared = true;
         }
     }
@@ -358,6 +364,9 @@ public final class ByteQuadsCanonicalizer
     /**********************************************************
      */
 
+    /**
+     * @return Number of symbol entries contained by this canonicalizer instance
+     */
     public int size()
     {
         if (_tableInfo != null) { // root table
@@ -368,7 +377,7 @@ public final class ByteQuadsCanonicalizer
     }
 
     /**
-     * Returns number of primary slots table has currently
+     * @return number of primary slots table has currently
      */
     public int bucketCount() { return _hashSize; }
 
@@ -376,6 +385,8 @@ public final class ByteQuadsCanonicalizer
      * Method called to check to quickly see if a child symbol table
      * may have gotten additional entries. Used for checking to see
      * if a child table should be merged into shared table.
+     *
+     * @return Whether main hash area has been modified
      */
     public boolean maybeDirty() { return !_hashShared; }
 
@@ -385,6 +396,8 @@ public final class ByteQuadsCanonicalizer
      * Method mostly needed by unit tests; calculates number of
      * entries that are in the primary slot set. These are
      * "perfect" entries, accessible with a single lookup
+     *
+     * @return Number of entries in the primary hash area
      */
     public int primaryCount()
     {
@@ -400,6 +413,8 @@ public final class ByteQuadsCanonicalizer
     /**
      * Method mostly needed by unit tests; calculates number of entries
      * in secondary buckets
+     *
+     * @return Number of entries in the secondary hash area
      */
     public int secondaryCount() {
         int count = 0;
@@ -415,6 +430,8 @@ public final class ByteQuadsCanonicalizer
     /**
      * Method mostly needed by unit tests; calculates number of entries
      * in tertiary buckets
+     *
+     * @return Number of entries in the tertiary hash area
      */
     public int tertiaryCount() {
         int count = 0;
@@ -429,7 +446,9 @@ public final class ByteQuadsCanonicalizer
 
     /**
      * Method mostly needed by unit tests; calculates number of entries
-     * in shared spillover area
+     * in shared spill-over area
+     *
+     * @return Number of entries in the linear spill-over areay
      */
     public int spilloverCount() {
         // difference between spillover end, start, divided by 4 (four ints per slot)
@@ -562,13 +581,16 @@ public final class ByteQuadsCanonicalizer
          * that contains actual quads.
          */
         if (qlen < 4) { // another sanity check
-            if (qlen == 3) {
+            switch (qlen) {
+            case 3:
                 return findName(q[0], q[1], q[2]);
-            }
-            if (qlen == 2) {
+            case 2:
                 return findName(q[0], q[1]);
+            case 1:
+                return findName(q[0]);
+            default: // if 0 ever passed
+                return "";
             }
-            return findName(q[0]);
         }
         final int hash = calcHash(q, qlen);
         int offset = _calcOffset(hash);
@@ -770,7 +792,6 @@ public final class ByteQuadsCanonicalizer
         _hashArea[offset+3] = 1;
         _names[offset >> 2] = name;
         ++_count;
-        _verifyNeedForRehash();
         return name;
     }
 
@@ -786,7 +807,6 @@ public final class ByteQuadsCanonicalizer
         _hashArea[offset+3] = 2;
         _names[offset >> 2] = name;
         ++_count;
-        _verifyNeedForRehash();
         return name;
     }
 
@@ -802,7 +822,6 @@ public final class ByteQuadsCanonicalizer
         _hashArea[offset+3] = 3;
         _names[offset >> 2] = name;
         ++_count;
-        _verifyNeedForRehash();
         return name;
     }
 
@@ -853,33 +872,15 @@ public final class ByteQuadsCanonicalizer
 
         // and finally; see if we really should rehash.
         ++_count;
-        _verifyNeedForRehash();
         return name;
     }
 
-    private void _verifyNeedForRehash() {
-        // Yes if above 80%, or above 50% AND have ~1% spill-overs
-        if (_count > (_hashSize >> 1)) { // over 50%
-            int spillCount = (_spilloverEnd - _spilloverStart()) >> 2;
-            if ((spillCount > (1 + _count >> 7))
-                    || (_count > (_hashSize * 0.80))) {
-                _needRehash = true;
-            }
-        }
-    }
-
     private void _verifySharing()
     {
         if (_hashShared) {
             _hashArea = Arrays.copyOf(_hashArea, _hashArea.length);
             _names = Arrays.copyOf(_names, _names.length);
             _hashShared = false;
-            // 09-Sep-2015, tatu: As per [jackson-core#216], also need to ensure
-            //    we rehash as needed, as need-rehash flag is not copied from parent
-            _verifyNeedForRehash();
-        }
-        if (_needRehash) {
-            rehash();
         }
     }
 
@@ -888,14 +889,20 @@ public final class ByteQuadsCanonicalizer
      */
     private int _findOffsetForAdd(int hash)
     {
-        // first, check the primary:
+        // first, check the primary: if slot found, no need for resize
         int offset = _calcOffset(hash);
         final int[] hashArea = _hashArea;
         if (hashArea[offset+3] == 0) {
 //System.err.printf(" PRImary slot #%d, hash %X\n", (offset>>2), hash & 0x7F);
             return offset;
         }
-        // then secondary
+
+        // Otherwise let's see if we are due resize():
+        if (_checkNeedForRehash()) {
+            return _resizeAndFindOffsetForAdd(hash);
+        }
+
+        // If not, proceed with secondary slot
         int offset2 = _secondaryStart + ((offset >> 3) << 2);
         if (hashArea[offset2+3] == 0) {
 //System.err.printf(" SECondary slot #%d (start x%X), hash %X\n",(offset >> 3), _secondaryStart, (hash & 0x7F));
@@ -929,13 +936,52 @@ public final class ByteQuadsCanonicalizer
             if (_failOnDoS) {
                 _reportTooManyCollisions();
             }
-            // and if we didn't fail, we'll simply force rehash for next add
-            // (which, in turn, may double up or nuke contents, depending on size etc)
-            _needRehash = true;
+            return _resizeAndFindOffsetForAdd(hash);
         }
         return offset;
     }
 
+    // @since 2.10
+    private int _resizeAndFindOffsetForAdd(int hash)
+    {
+        // First things first: we need to resize+rehash (or, if too big, nuke contents)
+        rehash();
+        
+        // Copy of main _findOffsetForAdd except for checks to resize: can not be needed
+        int offset = _calcOffset(hash);
+        final int[] hashArea = _hashArea;
+        if (hashArea[offset+3] == 0) {
+            return offset;
+        }
+        int offset2 = _secondaryStart + ((offset >> 3) << 2);
+        if (hashArea[offset2+3] == 0) {
+            return offset2;
+        }
+        offset2 = _tertiaryStart + ((offset >> (_tertiaryShift + 2)) << _tertiaryShift);
+        final int bucketSize = (1 << _tertiaryShift);
+        for (int end = offset2 + bucketSize; offset2 < end; offset2 += 4) {
+            if (hashArea[offset2+3] == 0) {
+                return offset2;
+            }
+        }
+        offset = _spilloverEnd;
+        _spilloverEnd += 4;
+        return offset;
+    }
+
+    // Helper method for checking if we should simply rehash() before add
+    private boolean _checkNeedForRehash() {
+        // Yes if above 80%, or above 50% AND have ~1% spill-overs
+        if (_count > (_hashSize >> 1)) { // over 50%
+            int spillCount = (_spilloverEnd - _spilloverStart()) >> 2;
+            if ((spillCount > (1 + _count >> 7))
+                    || (_count > (_hashSize * 0.80))) {
+                return true;
+            }
+        }
+        return false;
+    }
+    
     private int _appendLongName(int[] quads, int qlen)
     {
         int start = _longNameOffset;
@@ -1063,7 +1109,6 @@ public final class ByteQuadsCanonicalizer
 
     private void rehash()
     {
-        _needRehash = false;
         // Note: since we'll make copies, no need to unshare, can just mark as such:
         _hashShared = false;
 
