@@ -33,10 +33,10 @@ public final class ByteArrayBuilder extends OutputStream
     // Size of the first block we will allocate.
     private final static int INITIAL_BLOCK_SIZE = 500;
     
-    // Maximum block size we will use for individual non-aggregated
-    // blocks. Let's limit to using 256k chunks.
-    private final static int MAX_BLOCK_SIZE = (1 << 18);
-    
+    // Maximum block size we will use for individual non-aggregated blocks.
+    // For 2.10, let's limit to using 128k chunks (was 256k up to 2.9)
+    private final static int MAX_BLOCK_SIZE = (1 << 17);
+
     final static int DEFAULT_BLOCK_ARRAY_SIZE = 40;
 
     // Optional buffer recycler instance that we can use for allocating the first block.
@@ -54,9 +54,24 @@ public final class ByteArrayBuilder extends OutputStream
 
     public ByteArrayBuilder(BufferRecycler br, int firstBlockSize) {
         _bufferRecycler = br;
+        // 04-Sep-2020, tatu: Let's make this bit more robust and refuse to allocate
+        //    humongous blocks even if requested
+        if (firstBlockSize > MAX_BLOCK_SIZE) {
+            firstBlockSize = MAX_BLOCK_SIZE;
+        }
         _currBlock = (br == null) ? new byte[firstBlockSize] : br.allocByteBuffer(BufferRecycler.BYTE_WRITE_CONCAT_BUFFER);
     }
 
+    private ByteArrayBuilder(BufferRecycler br, byte[] initialBlock, int initialLen) {
+        _bufferRecycler = null;
+        _currBlock = initialBlock;
+        _currBlockPtr = initialLen;
+    }
+
+    public static ByteArrayBuilder fromInitial(byte[] initialBlock, int length) {
+        return new ByteArrayBuilder(null, initialBlock, length);
+    }
+    
     public void reset() {
         _pastLen = 0;
         _currBlockPtr = 0;
@@ -67,6 +82,8 @@ public final class ByteArrayBuilder extends OutputStream
     }
 
     /**
+     * @return Number of bytes aggregated so far
+     *
      * @since 2.9
      */
     public int size() {
@@ -115,9 +132,7 @@ public final class ByteArrayBuilder extends OutputStream
         }
     }
 
-    /**
-     * @since 2.9
-     */
+    // @since 2.9
     public void appendFourBytes(int b32) {
         if ((_currBlockPtr + 3) < _currBlock.length) {
             _currBlock[_currBlockPtr++] = (byte) (b32 >> 24);
@@ -135,6 +150,8 @@ public final class ByteArrayBuilder extends OutputStream
     /**
      * Method called when results are finalized and we can get the
      * full aggregated result buffer to return to the caller
+     *
+     * @return Aggregated contents as a {@code byte[]}
      */
     public byte[] toByteArray()
     {
@@ -172,6 +189,8 @@ public final class ByteArrayBuilder extends OutputStream
     /**
      * Method called when starting "manual" output: will clear out
      * current state and return the first segment buffer to fill
+     *
+     * @return Segment to use for writing
      */
     public byte[] resetAndGetFirstSegment() {
         reset();
@@ -182,6 +201,8 @@ public final class ByteArrayBuilder extends OutputStream
      * Method called when the current segment buffer is full; will
      * append to current contents, allocate a new segment buffer
      * and return it
+     *
+     * @return Segment to use for writing
      */
     public byte[] finishCurrentSegment() {
         _allocMore();
