@@ -180,6 +180,10 @@ public abstract class BaseTest
         if (!p.hasCurrentToken()) {
             p.nextToken();
         }
+        // first basic check, mostly for test coverage
+        assertNull(p.getTypeId());
+        assertNull(p.getObjectId());
+
         assertToken(JsonToken.START_OBJECT, p.currentToken()); // main object
 
         assertToken(JsonToken.FIELD_NAME, p.nextToken()); // 'Image'
@@ -302,7 +306,7 @@ public abstract class BaseTest
     
     /*
     /**********************************************************
-    /* Parser/generator construction
+    /* Parser construction
     /**********************************************************
      */
 
@@ -396,12 +400,34 @@ public abstract class BaseTest
         return f.createParser(input);
     }
 
+    /*
+    /**********************************************************
+    /* Generator construction
+    /**********************************************************
+     */
+
+    protected JsonGenerator createGenerator(OutputStream out) throws IOException {
+        return createGenerator(JSON_FACTORY, out);
+    }
+
+    protected JsonGenerator createGenerator(TokenStreamFactory f, OutputStream out) throws IOException {
+        return f.createGenerator(out);
+    }
+
+    protected JsonGenerator createGenerator(Writer w) throws IOException {
+        return createGenerator(JSON_FACTORY, w);
+    }
+
+    protected JsonGenerator createGenerator(TokenStreamFactory f, Writer w) throws IOException {
+        return f.createGenerator(w);
+    }
+
     /*
     /**********************************************************
     /* Helper read/write methods
     /**********************************************************
      */
-    
+
     protected void writeJsonDoc(JsonFactory f, String doc, Writer w) throws IOException
     {
         writeJsonDoc(f, doc, f.createGenerator(w));
@@ -566,7 +592,12 @@ public abstract class BaseTest
     protected JsonFactory newStreamFactory() {
         return new JsonFactory();
     }
-    
+
+    // @since 2.9.8
+    protected JsonFactoryBuilder streamFactoryBuilder() {
+        return (JsonFactoryBuilder) JsonFactory.builder();
+    }
+
     protected String fieldNameFor(int index)
     {
         StringBuilder sb = new StringBuilder(16);
