@@ -1,12 +1,16 @@
 package com.fasterxml.jackson.core.read;
 
 import java.io.ByteArrayInputStream;
+import java.io.CharArrayReader;
 import java.io.IOException;
 import java.io.StringReader;
 import java.math.BigDecimal;
 import java.math.BigInteger;
 
 import com.fasterxml.jackson.core.*;
+import com.fasterxml.jackson.core.exc.InputCoercionException;
+import com.fasterxml.jackson.core.io.NumberInput;
+import com.fasterxml.jackson.core.json.JsonReadFeature;
 
 /**
  * Set of basic unit tests for verifying that the basic parser
@@ -16,7 +20,13 @@ import com.fasterxml.jackson.core.*;
 public class NumberParsingTest
     extends com.fasterxml.jackson.core.BaseTest
 {
-    private final JsonFactory FACTORY = new JsonFactory();
+    private final JsonFactory FACTORY = sharedStreamFactory();
+
+    /*
+    /**********************************************************************
+    /* Tests, Boolean
+    /**********************************************************************
+     */
 
     public void testSimpleBoolean() throws Exception
     {
@@ -36,9 +46,15 @@ public class NumberParsingTest
         p.close();
     }
 
+    /*
+    /**********************************************************************
+    /* Tests, Int
+    /**********************************************************************
+     */
+
     public void testSimpleInt() throws Exception
     {
-        for (int EXP_I : new int[] { 1234, -999, 0, 1, -2 }) {
+        for (int EXP_I : new int[] { 1234, -999, 0, 1, -2, 123456789 }) {
             _testSimpleInt(EXP_I, MODE_INPUT_STREAM);
             _testSimpleInt(EXP_I, MODE_INPUT_STREAM_THROTTLED);
             _testSimpleInt(EXP_I, MODE_READER);
@@ -53,8 +69,34 @@ public class NumberParsingTest
         assertToken(JsonToken.START_ARRAY, p.nextToken());
         assertToken(JsonToken.VALUE_NUMBER_INT, p.nextToken());
         assertEquals(JsonParser.NumberType.INT, p.getNumberType());
+        assertTrue(p.isExpectedNumberIntToken());
         assertEquals(""+EXP_I, p.getText());
 
+        if (((short) EXP_I) == EXP_I) {
+            assertEquals((short) EXP_I, p.getShortValue());
+            if (((byte) EXP_I) == EXP_I) {
+                assertEquals((byte) EXP_I, p.getByteValue());
+            } else {
+                // verify overflow
+                try {
+                    p.getByteValue();
+                    fail("Should get exception for non-byte value "+EXP_I);
+                } catch (InputCoercionException e) {
+                    verifyException(e, "Numeric value");
+                    verifyException(e, "out of range");
+                }
+            }
+        } else {
+            // verify overflow
+            try {
+                p.getShortValue();
+                fail("Should get exception for non-short value "+EXP_I);
+            } catch (InputCoercionException e) {
+                verifyException(e, "Numeric value");
+                verifyException(e, "out of range");
+            }
+        }
+
         assertEquals(EXP_I, p.getIntValue());
         assertEquals(EXP_I, p.getValueAsInt(EXP_I + 3));
         assertEquals(EXP_I, p.getValueAsInt());
@@ -67,20 +109,29 @@ public class NumberParsingTest
         DOC = String.valueOf(EXP_I);
         p = createParser(mode, DOC + " "); // DataInput requires separator
         assertToken(JsonToken.VALUE_NUMBER_INT, p.nextToken());
+        assertTrue(p.isExpectedNumberIntToken());
         assertEquals(DOC, p.getText());
 
-        int i = 0;
-        
-        try {
-            i = p.getIntValue();
-        } catch (Exception e) {
-            throw new Exception("Failed to parse input '"+DOC+"' (parser of type "+p.getClass().getSimpleName()+")", e);
-        }
+        int i = p.getIntValue();
+
         assertEquals(EXP_I, i);
         assertEquals((long) EXP_I, p.getLongValue());
         assertEquals((double) EXP_I, p.getDoubleValue());
         assertEquals(BigDecimal.valueOf((long) EXP_I), p.getDecimalValue());
         p.close();
+
+        // and finally, coercion from double to int; couple of variants
+        DOC = String.valueOf(EXP_I)+".0";
+        p = createParser(mode, DOC + " ");
+        assertToken(JsonToken.VALUE_NUMBER_FLOAT, p.nextToken());
+        assertFalse(p.isExpectedNumberIntToken());
+        assertEquals(EXP_I, p.getValueAsInt());
+        p.close();
+
+        p = createParser(mode, DOC + " ");
+        assertToken(JsonToken.VALUE_NUMBER_FLOAT, p.nextToken());
+        assertEquals(EXP_I, p.getValueAsInt(0));
+        p.close();
     }
 
     public void testIntRange() throws Exception
@@ -101,6 +152,49 @@ public class NumberParsingTest
         }
     }
 
+    public void testIntParsing() throws Exception
+    {
+        char[] testChars = "123456789".toCharArray();
+
+        assertEquals(3, NumberInput.parseInt(testChars, 2, 1));
+        assertEquals(123, NumberInput.parseInt(testChars, 0, 3));
+        assertEquals(2345, NumberInput.parseInt(testChars, 1, 4));
+        assertEquals(9, NumberInput.parseInt(testChars, 8, 1));
+        assertEquals(456789, NumberInput.parseInt(testChars, 3, 6));
+        assertEquals(23456, NumberInput.parseInt(testChars, 1, 5));
+        assertEquals(123456789, NumberInput.parseInt(testChars, 0, 9));
+
+        testChars = "32".toCharArray();
+        assertEquals(32, NumberInput.parseInt(testChars, 0, 2));
+        testChars = "189".toCharArray();
+        assertEquals(189, NumberInput.parseInt(testChars, 0, 3));
+
+        testChars = "10".toCharArray();
+        assertEquals(10, NumberInput.parseInt(testChars, 0, 2));
+        assertEquals(0, NumberInput.parseInt(testChars, 1, 1));
+    }
+
+    public void testIntParsingWithStrings() throws Exception
+    {
+        assertEquals(3, NumberInput.parseInt("3"));
+        assertEquals(0, NumberInput.parseInt("0"));
+        assertEquals(-3, NumberInput.parseInt("-3"));
+        assertEquals(27, NumberInput.parseInt("27"));
+        assertEquals(-31, NumberInput.parseInt("-31"));
+        assertEquals(271, NumberInput.parseInt("271"));
+        assertEquals(-131, NumberInput.parseInt("-131"));
+        assertEquals(2709, NumberInput.parseInt("2709"));
+        assertEquals(-9999, NumberInput.parseInt("-9999"));
+        assertEquals(Integer.MIN_VALUE, NumberInput.parseInt(""+Integer.MIN_VALUE));
+        assertEquals(Integer.MAX_VALUE, NumberInput.parseInt(""+Integer.MAX_VALUE));
+    }
+
+    /*
+    /**********************************************************************
+    /* Tests, Long
+    /**********************************************************************
+     */
+
     public void testSimpleLong() throws Exception
     {
         _testSimpleLong(MODE_INPUT_STREAM);
@@ -124,8 +218,10 @@ public class NumberParsingTest
         // Should get an exception if trying to convert to int 
         try {
             p.getIntValue();
-        } catch (JsonParseException pe) {
-            verifyException(pe, "out of range");
+        } catch (InputCoercionException e) {
+            verifyException(e, "out of range");
+            assertEquals(JsonToken.VALUE_NUMBER_INT, e.getInputType());
+            assertEquals(Integer.TYPE, e.getTargetType());
         }
         assertEquals((double) EXP_L, p.getDoubleValue());
         assertEquals(BigDecimal.valueOf((long) EXP_L), p.getDecimalValue());
@@ -162,6 +258,55 @@ public class NumberParsingTest
         }
     }
 
+    public void testLongParsing() throws Exception
+    {
+        char[] testChars = "123456789012345678".toCharArray();
+
+        assertEquals(123456789012345678L, NumberInput.parseLong(testChars, 0, testChars.length));
+    }
+
+    public void testLongBoundsChecks() throws Exception
+    {
+        String minLong = String.valueOf(Long.MIN_VALUE).substring(1);
+        String belowMinLong = BigInteger.valueOf(Long.MIN_VALUE).subtract(BigInteger.ONE)
+                .toString().substring(1);
+        String maxLong = String.valueOf(Long.MAX_VALUE);
+        String aboveMaxLong = BigInteger.valueOf(Long.MAX_VALUE).add(BigInteger.ONE).toString();
+        final String VALUE_491 = "1323372036854775807"; // is within range (JACKSON-491)
+        final String OVERFLOW =  "9999999999999999999"; // and this one is clearly out
+
+        assertTrue(NumberInput.inLongRange(minLong, true));
+        assertTrue(NumberInput.inLongRange(maxLong, false));
+        assertTrue(NumberInput.inLongRange(VALUE_491, true));
+        assertTrue(NumberInput.inLongRange(VALUE_491, false));
+
+        assertFalse(NumberInput.inLongRange(OVERFLOW, false));
+        assertFalse(NumberInput.inLongRange(OVERFLOW, true));
+        assertFalse(NumberInput.inLongRange(belowMinLong, true));
+        assertFalse(NumberInput.inLongRange(aboveMaxLong, false));
+
+        char[] cbuf = minLong.toCharArray();
+        assertTrue(NumberInput.inLongRange(cbuf, 0, cbuf.length, true));
+        cbuf = maxLong.toCharArray();
+        assertTrue(NumberInput.inLongRange(cbuf, 0, cbuf.length, false));
+        cbuf = VALUE_491.toCharArray();
+        assertTrue(NumberInput.inLongRange(cbuf, 0, cbuf.length, true));
+        assertTrue(NumberInput.inLongRange(cbuf, 0, cbuf.length, false));
+
+        cbuf = OVERFLOW.toCharArray();
+        assertFalse(NumberInput.inLongRange(cbuf, 0, cbuf.length, true));
+        assertFalse(NumberInput.inLongRange(cbuf, 0, cbuf.length, false));
+
+        cbuf = aboveMaxLong.toCharArray();
+        assertFalse(NumberInput.inLongRange(cbuf, 0, cbuf.length, false));
+    }
+    
+    /*
+    /**********************************************************************
+    /* Tests, BigXxx
+    /**********************************************************************
+     */
+    
     public void testBigDecimalRange() throws Exception
     {
         for (int mode : ALL_MODES) {
@@ -203,6 +348,12 @@ public class NumberParsingTest
             p.close();
         }
     }
+
+    /*
+    /**********************************************************************
+    /* Tests, floating point (basic)
+    /**********************************************************************
+     */
     
     public void testSimpleDouble() throws Exception
     {
@@ -252,6 +403,46 @@ public class NumberParsingTest
         }
     }
 
+    public void testFloatBoundary146Chars() throws Exception
+    {
+        final char[] arr = new char[50005];
+        for(int i = 500; i != 9000; ++i) {
+          java.util.Arrays.fill(arr, 0, i, ' ');
+          arr[i] = '-';
+          arr[i + 1] = '1';
+          arr[i + 2] = 'e';
+          arr[i + 3] = '-';
+          arr[i + 4] = '1';
+          CharArrayReader r = new CharArrayReader(arr, 0, i+5);
+          JsonParser p = FACTORY.createParser(r);
+          assertToken(JsonToken.VALUE_NUMBER_FLOAT, p.nextToken());
+          p.close();
+        }
+    }
+
+    public void testFloatBoundary146Bytes() throws Exception
+    {
+        final byte[] arr = new byte[50005];
+        for(int i = 500; i != 9000; ++i) {
+            java.util.Arrays.fill(arr, 0, i, (byte) 0x20);
+            arr[i] = '-';
+            arr[i + 1] = '1';
+            arr[i + 2] = 'e';
+            arr[i + 3] = '-';
+            arr[i + 4] = '1';
+            ByteArrayInputStream in = new ByteArrayInputStream(arr, 0, i+5);
+            JsonParser p = FACTORY.createParser(in);
+            assertToken(JsonToken.VALUE_NUMBER_FLOAT, p.nextToken());
+            p.close();
+        }
+    }
+
+    /*
+    /**********************************************************************
+    /* Tests, misc other
+    /**********************************************************************
+     */
+
     public void testNumbers() throws Exception
     {
         _testNumbers(MODE_INPUT_STREAM);
@@ -428,8 +619,9 @@ public class NumberParsingTest
      */
     public void testParsingOfLongerSequencesWithNonNumeric() throws Exception
     {
-        JsonFactory f = new JsonFactory();
-        f.enable(JsonParser.Feature.ALLOW_NON_NUMERIC_NUMBERS);
+        JsonFactory f = JsonFactory.builder()
+                .enable(JsonReadFeature.ALLOW_NON_NUMERIC_NUMBERS)
+                .build();
         _testParsingOfLongerSequencesWithNonNumeric(f, MODE_INPUT_STREAM);
         _testParsingOfLongerSequencesWithNonNumeric(f, MODE_INPUT_STREAM_THROTTLED);
         _testParsingOfLongerSequencesWithNonNumeric(f, MODE_READER);
